-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}
M.hydros = {}

local function nop()
end

local function updateGFX(dt)
    -- update the source command value
    for i,h in pairs(M.hydros) do
        h.cmd = math.min(math.max(electrics.values[h.inputSource] or 0, h.inputInLimit), h.inputOutLimit) * h.inputFactor
        if h.cmd ~= h.inputCenter or h.analogue == true then
            h._inrate = h.inRate
            h._outrate = h.outRate
        else
            -- set autocenter rate
            h._inrate = h.autoCenterRate
            h._outrate = h.autoCenterRate
        end

        if h.cmd >= h.inputCenter then
            h.cmd = h.cOut + h.cmd * h.multOut
        else
            h.cmd = h.cIn + h.cmd * h.multIn
        end

        if h.state > h.center then
            h.hydroDirState = (h.state - h.center) * h.invMultOut * h.inputFactorSign
        else
            h.hydroDirState = (h.state - h.center) * h.invMultIn * h.inputFactorSign
        end
    end
end

local function update(dts)
    -- state: the state of the hydro from -1 to 1
    -- cmd the input value
    -- note: state is scaled to the ratio as the last step
    for i,h in pairs(M.hydros) do
        -- slowly approach the desired value
        if h.cmd < h.state then
            h.state = math.max(h.state - dts * h._inrate, h.cmd)
        else
            h.state = math.min(h.state + dts * h._outrate, h.cmd)
        end

        --if inputSource ~= "steering" then
        --    print("state: "..h.state .. " - inlimit: " .. h.inLimit.. " - state: "..h.state .. " cmd:"..h.cmd .. " val = "..val)
        --end
        obj:setBeamLengthRefRatio(h.beam.cid, h.state)
    end
end

-- nop'ed functions
M.updateGFX = updateGFX
M.update = update

local function init()
    if v.data.hydros == nil then
        M.updateGFX = nop
        M.update = nop
        M.hydros = {}
        return
    end

    M.hydros = shallowcopy(v.data.hydros)

    local hydroCount = 0
    for i, h in pairs (M.hydros) do
        hydroCount = hydroCount + 1
        h.inputCenter = h.inputCenter * h.inputFactor
        h.inputInLimit = h.inputInLimit * h.inputFactor
        h.inputOutLimit = h.inputOutLimit * h.inputFactor
        h.inputFactorSign = sign(h.inputFactor)

        if h.inputFactor < 0 then
            h.inputInLimit, h.inputOutLimit = h.inputOutLimit, h.inputInLimit
        end

        local inputMiddle = (h.inputOutLimit + h.inputInLimit) * 0.5
        if h.inputCenter >= inputMiddle then
            h.center = 1 + (h.outLimit - 1) * (h.inputCenter - inputMiddle) / (h.inputOutLimit - inputMiddle)
        else
            h.center = 1 - (1 - h.inLimit) * (inputMiddle - h.inputCenter) / (inputMiddle - h.inputInLimit)
        end
        h.state = h.center
        h.multOut = (h.outLimit - h.center) / (h.inputOutLimit - h.inputCenter)
        h.cOut = h.center - h.inputCenter * h.multOut
        h.multIn = (h.center - h.inLimit) / (h.inputCenter - h.inputInLimit)
        h.cIn = h.center - h.inputCenter * h.multIn
        h.cmd = h.inputCenter
        h.invMultOut = 1 / (h.outLimit - h.center)
        h.invMultIn = 1 / (h.center - h.inLimit)
        h._inrate = h.inRate
        h._outrate = h.outRate
        h.hydroDirState = 0
        
        -- little workaround
        if h.inputSource == "steering" then
            h.inputSource = "steering_input"
        end
    end

    if hydroCount == 0 then
        M.updateGFX = nop
        M.update = nop
    end
end

local function reset()
    for k,h in pairs(M.hydros) do
        h.state = h.center
        h.cmd = h.inputCenter
        h._inrate = h.inRate
        h._outrate = h.outRate
    end
end

local function sendHydroStateToGUI()
    gameEngine:executeJS("HookManager.trigger('HydrosUpdate', "..encodeJson(M.hydros)..");")
end

-- public interface
M.init = init
M.reset = reset
M.sendHydroStateToGUI = sendHydroStateToGUI

return M