-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}

M.active = false


local raw = {}
local c = nil
local ct = nil
local debugDrawer = nil

local mode = 0
local mode2 = 0
local deviceMode = ""
local deviceName = ""


local result = {}

local x = 50
local y = 300
local pFontLeft = nil
local pBGColor = nil

local function showGUI()
	local g = string.format([[beamngguiconfig 1
callback activeVehicle inputwizard.guiCallback
title Input Configuration [%d]

container
  type = verticalStack
  name = root

]], mode)

	if mode == 0 then
		g = g .. [[
control
	type = button
	name = wheel
	description = Steering Wheel
	level = 1

control
	type = button
	name = joystick
	description = Joystick
	level = 1

control
	type = button
	name = stickshift
	description = Stickshift
	level = 1

control
	type = button
	name = gamepad
	description = Gamepad
	level = 1

control
	type = button
	name = utilkeys
	description = Utility Keys
	level = 1

]]

	elseif mode == 1 then
		--dump(input.rawDevices)
		for k,d in pairs(input.rawDevices) do
			g = g .. string.format([[
control
  type = button
  name = %s
  description = %s
  level = 1

]], k, d.product_name)
		end
		g = g .. [[
control
	type = button
	name = back
	description = Back
	level = 1

]]
	elseif mode == 2 then
		g = g .. [[
control
	type = button
	name = next
	description = Next
	level = 1

control
	type = button
	name = previous
	description = Previous
	level = 1
	
control
	type = button
	name = back
	description = Restart
	level = 1
	
control
	type = button
	name = done
	description = Done
	level = 1

]]
	end
	--print(g)
	gameEngine:showGUI(g)
end


local function activate()
	if not gameEngine then return end
	if not debugDrawer then debugDrawer = obj:getDebugDrawProxy() end
	
	debugDrawer:showDebug(true)
	ct = obj.canvasTexture
	if not ct then return end
	c = ct:getCanvas()
	-- init gui things
	pFontLeft = SkPaint()
	pFontLeft:setAntiAlias(true)
	pFontLeft:setColor(SK_ColorBLACK)
	pFontLeft:setTextAlign(kLeft_Align)
	pFontLeft:setTextSize(25)
	pFontLeft:setTypeface(SkTypeface.CreateFromName("Verdana", kNormal))
	
	pBGColor = SkPaint()
	pBGColor:setAntiAlias(true)
	pBGColor:setARGB(180,255,255,255)
	-- end
	
	M.active = true
	gameEngine:enableInputEventForwarding("inputwizard.processRawEvent")
	showGUI()
end

local function finish()
	M.active = false
	debugDrawer:showDebug(false)
	gameEngine:enableInputEventForwarding("")
end

local function detectAxis()
	if raw[deviceName] then
		for k1,objtype in pairs(raw[deviceName]) do
			for k2,obj in pairs(objtype) do
				local txt = k1.."." .. k2 .. " = " .. string.format("%0.3f", obj.fValue * 100)
				c:drawText(txt, string.len(txt), x, y, pFontLeft)
				y = y + 40
				if obj.fValue > 0.95 then
					return {objtype=k1,inst=k2,min=0,max=1}
				end
			end
		end
	end
	return nil
end

local function updateCanvas(dt)
	ct:clear()
	c:drawRect(ct.rect, pBGColor)
	
	x = 50
	y = 300
	
	if mode == 2 then
		if deviceMode == "wheel" then
			if mode2 == 0 then
				local txt = "Please center your steering wheel and press next"
				c:drawText(txt, string.len(txt), x, y, pFontLeft)
				raw = {}
			elseif mode2 == 1 then
				local txt = "Move the steering wheel to the left until the end"
				c:drawText(txt, string.len(txt), x, y, pFontLeft)
				y = y + 40
				local res = detectAxis()
				if res then
					result.steerleft = res
					mode2 = mode2 + 1
					raw = {}
					return
				end
			elseif mode2 == 2 then
				local txt = "Move the steering wheel to the right until the end"
				c:drawText(txt, string.len(txt), x, y, pFontLeft)
				y = y + 40
				local res = detectAxis()
				if res then
					result.steerright = res
					mode2 = mode2 + 1
					raw = {}
					return
				end
			end
		end
	end
end


local function update(dt)
	--print("inputwizard.update()")


	updateCanvas(dt)
end


local function guiCallback(guiMode, str)
	print("inputwizard.guiCallback("..guiMode..","..str..")")
	if guiMode == "button" then
		local args = unserialize(str)
		local btnName = args.button
		--print(btnName)
		if btnName == "Back" then
			-- back to main menu
			mode = 0
			mode2 = 0
			showGUI()
			return
		end
		if mode == 0 then
			deviceMode = btnName
			mode = 1
			showGUI()
			return
		elseif mode == 1 then
			deviceName = btnName
			mode = 2
			showGUI()
			return
		elseif mode == 2 and btnName == "Previous" then
			mode2 = mode2 - 1
			if mode2 < 0 then mode2 = 0 end
			return
		elseif mode == 2 and btnName == "next" then
			mode2 = mode2 + 1
			return
		end
	end
end

-- deviceInst : Device instance: joystick0, joystick1, etc
-- fValue : Value typically ranges from -1.0 to 1.0, but doesn't have to. - It depends on the context.
-- fValue2, fValue3, fValue4 : Extended float values (often used for absolute rotation Quat)
-- iValue : Signed integer value
-- action : InputActionType
-- deviceType : InputDeviceTypes
-- objType : InputEventType
-- objInst : InputObjectInstances
-- ascii : ASCII character code if this is a keyboard event.
-- modifier : Modifiers to action: SI_LSHIFT, SI_LCTRL, etc.
local function processRawEvent(deviceInst, fValue, fValue2, fValue3, fValue4, iValue, action, deviceType, objType, objInst, ascii, modifier)
	print("InputEvent("..deviceInst..", "..fValue..", "..fValue2..", "..fValue3..", "..fValue4..", "..iValue..", "..action..", "..deviceType..", "..objType..", "..objInst..", "..ascii..", "..modifier..")")

	local dev = deviceType..deviceInst

	if raw[dev] == nil then raw[dev] = {} end
	if raw[dev][objType] == nil then raw[dev][objType] = {} end
	if raw[dev][objType][objInst] == nil then raw[dev][objType][objInst] = {} end


	local d = raw[dev][objType][objInst]

	d.fValue  = fValue
	d.fValue2 = fValue2
	d.fValue3 = fValue3
	d.fValue4 = fValue4
	d.iValue  = iValue
	d.action  = action

	--dump(raw)
end

-- public interface
M.reset = reset
M.updateGFX = updateGFX
M.drawCanvas = drawCanvas
M.guiCallback = guiCallback
M.activate = activate
M.processRawEvent = processRawEvent
M.update = update

return M