-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}

M.config = {}

local partsChanged = false

local function save(fn)
    local file, err = io.open(fn, "w")
    if file then
        local s = serialize(v.userPartConfig)
        file:write(s)
        file:close()
    else
        logError("unable to open file for writing: "..fn)
    end
end

local function load(fn, guiMode, reload)
    local file, err = io.open(fn, "r")
    if file then
        local content = file:read("*all")
        file:close()
        local args = unserialize(content)
        args.guiMode = guiMode
        setConfig(args, reload)
    else
        logError("unable to open file for reading: "..fn)
    end
end

function setConfig(args, reload)
    if tableSize(args) == 0 then return end

    -- TODO: fix it up properly that it does not require a reset
    -- the persistence-data is destroyed on the reset of this reset then which breaks the persistency :(
    M.config = args
    if reload == nil then reload = true end
    v.userPartConfig = args
    if reload then
        obj:requestReset(RESET_COMPLETELY)
        obj:requestReset(RESET_PHYSICS)
        partsChanged = true
    end
end

local function showGUI()
    local g = [[beamngguiconfig 1
callback activeVehicle partmgmt.guiCallback
title Parts Configuration

container
  type = verticalStack
  name = root
  height = 580

container
  type = scrolledVerticalSection
  description = Part config
  parent = root

]]

    -- horrible hack to find common string
    local commonString = ''
    for i=1, 30 do
        local cs = string.sub(v.slotDescriptions[next(v.slotDescriptions)], 1, i)
        local ok = true
        for k,va in v.slotMap:opairs() do
            local lcs = string.sub(v.slotDescriptions[k], 1, i)
            --print("'"..lcs .. "' == '" .. cs.. "'")
            if lcs ~= cs then
                ok = false
                break
            end
        end
        if not ok then
            break
        end
        commonString = cs
    end
    --print("#############: '"..commonString.."'")

    --local sm = table.sort(v.slotMap)
    --dump(v.slotMap)
    --local m = 0
    for k,va in v.slotMap:opairs() do
        g = g .. "control\n"
        g = g .. "  type = chooser\n"
        g = g .. "  name = " .. k .. "\n"

        local desc = v.slotDescriptions[k]
        if string.sub(desc, 1, string.len(commonString)) == commonString then
            desc = string.sub(desc, string.len(commonString))
        end
        g = g .. "  description = " .. desc .. "\n"

        local level = -1
        -- now enumerate the options
        local selection = false
        for k2,v2 in pairs(va) do

            local name = v2.name
            if string.sub(v2.name, 1, string.len(commonString)) == commonString then
                name = string.sub(name, string.len(commonString))
            end

            g = g .. "  option = " .. v2.partName .. " " .. name .. "\n"
            if v2.active == true then
                g = g .. "  selection = " .. name .. "\n"
                selection = true
            end
            level = v2.level
        end
        g = g .. "  level = " .. level .. "\n"
        g = g .. "  option = none Empty\n"
        if selection == false then
            g = g .. "  selection = Empty\n"
        end

        g = g .. "\n"
        --m = m + 1
        --if m > 3 then break end
    end
    -- and some additional things

    -- reset button
    g = g .. [[

container
  type = scrolledVerticalSection
  description = Save / Load / Reset
  parent = root

control
  type = saveButton
  name = saveBtn
  icon = tools/gui/images/save-icon_d.png
  filefilter = BeamNG Part Configurations Files (*.pc)|*.pc|All Files (*.*)|*.*|
  defaultpath = ]] .. v.vehicleDirectory .. [[
  defaultfileext = .pc
  defaultfilename = ]] .. v.vehicleDirectory .. [[part.pc
  description = Save
  level = 3

control
  type = loadButton
  name = loadBtn
  icon = tools/gui/images/open-file_d.png
  filefilter = BeamNG Part Configurations Files (*.pc)|*.pc|All Files (*.*)|*.*|
  defaultpath = ]] .. v.vehicleDirectory .. [[
  defaultfileext = .pc
  defaultfilename = ]] .. v.vehicleDirectory .. [[part.pc
  description = Load
  level = 3

control
  type = button
  name = resetBtn
  icon = tools/gui/images/iconCancel.png
  description = Reset
  level = 3

control
  type = doneButton
  icon = tools/gui/images/iconAccept.png
  description = Done
  level = 2

]]
    --print(g)
    gameEngine:showGUI(g)
end


local function doPartChanges()
    v:doPartChanges(obj)
    if v.userPartConfig.guiMode == 1 then
        showGUI()
    end
end

local function destroyed()
    if partsChanged == true then
        doPartChanges(obj)
        partsChanged = false
        return true
    end
    return false
end

local function selectPart(partName)
    -- only show that mesh
    obj.mesh:setAlpha(0.3, "", true)
    for flexKey, flexbody in pairs (v.data.flexbodies) do
        if flexbody.partOrigin == partName then
            obj.mesh:setAlpha(1, flexbody.mesh, false)
        end
    end
    if v.data.props then
	    for propKey, prop in pairs (v.data.props) do
            if prop.partOrigin == partName then
                obj.mesh:setAlpha(1, prop.mesh, false)
            end
        end
    end
end

local function selectReset()
    -- show all
    obj.mesh:setAlpha(1, "", false)
end



local function guiCallback(mode, str)
    --print("guiCallback("..tostring(mode)..","..tostring(str)..")")
    if not obj.mesh then return end
    if mode == "button" then
        local args = unserialize(str)
        local btnName = args.button
        if btnName == "resetBtn" then
            setConfig({})
        elseif btnName == "saveBtn" then
            local fn = args.saveBtn
            -- clean up the args
            args.button = nil
            args.saveBtn = nil
            args.loadBtn = nil
            args.resetBtn = nil
            -- set the config and then save
            setConfig(args, false)
            save(fn)
        elseif btnName == "loadBtn" then
            selectReset()
            load(args.loadBtn, 1)
        end
    elseif mode == "apply" then
        local args = unserialize(str)
        if args.DebugMode ~= nil then
            --bdebug.setMode(tonumber(args.DebugMode))
        end
        args.guiMode = 1
        if args.done ~= nil then args.guiMode = 0 end
        --dump(args)
        setConfig(args)
    elseif mode == "selectStart" then
        selectPart(str)
    elseif mode == "selectEnd" then
        selectReset()
    end
end

local function onDeserialized()
    --print("partmgmt.onDeserialized()")
    setConfig(M.config)
end


local function resetConfig()
    setConfig({})
end

-- public interface
M.destroyed = destroyed
M.save = save
M.load = load
M.selectPart = selectPart
M.selectReset = selectReset
M.showGUI = showGUI
M.guiCallback = guiCallback
M.setConfig = setConfig
M.onDeserialized = onDeserialized
M.resetConfig = resetConfig

return M